from pathlib import Path
import tempfile

import pytest

from mloq.api import setup_repository
from mloq.tests.config.test_generation import project_config, template


@pytest.fixture()
def repo_path():
    return Path(__file__).parent / "test_project"


def test_init_repository(project_config, template):
    with tempfile.TemporaryDirectory() as tmp:
        setup_repository(
            path=Path(tmp), project_config=project_config, template=template, override=False
        )



"""Contains the functions that generate the required configuration."""
from typing import Optional

from mloq.config.logic import get_docker_image, load_empty_config
from mloq.config.params import Config, PROJECT_CONFIG, TEMPLATE


def generate_project_config(
    project_config: Optional[Config] = None, interactive: bool = False
) -> Config:
    """
    Generate a dictionary containing the necessary values to define the project configuration.

    The generated config contains all the values defined under the project_config \
    section in mloq.yml.

    Args:
        project_config: dict containing the default values for the different parameters parsed.
        interactive: If True, the non-defined values can be provided from the CLI. \
                     If False, an error will be raised if a parameter is not defined \
                     as an environment variable, or it's present on the provided project dict.

    Returns:
        Dictionary containing the generated project_config parameters.
    """
    _project = project_config or {}
    project_config = load_empty_config()["project_config"]
    project_config.update(_project)
    # Fill in project values
    project_config["open_source"] = PROJECT_CONFIG["open_source"](
        project_config, interactive, default=True
    )
    # use_docker = PROJECT_CONFIG["docker"](project_config, interactive, default=True)
    project_config["docker"] = True  # use_docker
    project_config["ci"] = PROJECT_CONFIG["ci"](project_config, interactive, default="python")
    project_config["mlflow"] = PROJECT_CONFIG["mlflow"](project_config, interactive, default=False)
    project_config["requirements"] = PROJECT_CONFIG["requirements"](
        project_config, interactive, default="None"
    )
    return project_config


def generate_template(
    template: Optional[Config] = None,
    project_config: Optional[Config] = None,
    interactive: bool = False,
) -> Config:
    """
    Generate a dictionary containing the necessary values to customize the files generate by mloq.

    The generated config contains all the values defined under the template \
    section in mloq.yml.

    Args:
        template: dict containing the default values for the different template parameters parsed.
        project_config: dict containing the default values for the project_config parameters.
        interactive: If True, the non-defined values can be provided from the CLI. \
                     If False, an error will be raised if a parameter is not defined \
                     as an environment variable, or it's present on the provided project dict.

    Returns:
        Dictionary containing the generated template parameters.
    """
    # Initialize configuration placeholders
    _template, _project = template or {}, project_config or {}
    config = load_empty_config()
    project_config, template = config["project_config"], config["template"]
    project_config.update(_project)
    template.update(_template)
    # Fill in template values
    template["project_name"] = TEMPLATE["project_name"](template, interactive)
    template["description"] = TEMPLATE["description"](template, interactive)
    template["default_branch"] = TEMPLATE["default_branch"](
        template, interactive, default="master"
    )
    template["owner"] = TEMPLATE["owner"](template, interactive)
    template["author"] = TEMPLATE["author"](template, interactive, default=template["owner"])
    template["email"] = TEMPLATE["email"](template, interactive)
    copyright_holder = TEMPLATE["copyright_holder"](
        template, interactive, default=template["owner"]
    )
    template["copyright_holder"] = copyright_holder
    default_url = f"https://github.com/{template['owner']}/{template['project_name']}"
    template["project_url"] = TEMPLATE["project_url"](template, interactive, default=default_url)
    versions = TEMPLATE["python_versions"](template, interactive, default="3.6, 3.7, 3.8, 3.9")
    template["python_versions"] = versions
    # If project is undefined parse the full template
    if project_config["ci"] or project_config["ci"] is None:
        template["bot_name"] = TEMPLATE["bot_name"](
            template, interactive, default=template["author"]
        )
        template["bot_email"] = TEMPLATE["bot_email"](
            template, interactive, default=template["email"]
        )
    if not project_config["open_source"] or project_config["open_source"] is None:
        template["license"] = "proprietary"
    else:
        template["license"] = TEMPLATE["license_type"](template, interactive, default="MIT")
    base_docker = get_docker_image(template=template, project_config=project_config)
    if base_docker is not None:
        base_docker = TEMPLATE["docker_image"](template, interactive, default=base_docker)
    template["docker_image"] = str(base_docker) if base_docker is None else base_docker
    return template



"""This module defines the functionality to set up Github Actions workflows programmatically."""
from pathlib import Path
from typing import Union

from mloq import _logger
from mloq.config.params import Config, is_empty
from mloq.directories import create_github_actions_directories
from mloq.files import push_dist_wkf, push_python_wkf
from mloq.templating import write_template


WORKFLOW_NAMES = {
    "dist": push_dist_wkf,
    "python": push_python_wkf,
}


def setup_workflow_template(
    root_path: Union[Path, str], project_config: Config, template: Config, override: bool = False
):
    """Add the target workflows to the corresponding .github/workflows repository."""
    workflow = project_config.get("ci", "empty")
    if is_empty(workflow):
        return
    root_path = Path(root_path)
    create_github_actions_directories(root_path)
    workflows_path = Path(root_path) / ".github" / "workflows"
    # TODO: Check for incompatible workflows
    workflow_file = WORKFLOW_NAMES.get(workflow)
    if workflow_file is None:
        _logger.warning(f"Workflow {workflow} not defined. Skipping")
    else:
        write_template(workflow_file, template=template, path=workflows_path, override=override)


def setup_push_workflow(
    project_config: Config, template: Config, path: Union[str, Path], override: bool = False
):
    """Initialize the target workflow."""
    if is_empty(project_config.get("ci", "empty")):
        return
    create_github_actions_directories(path)
    setup_workflow_template(
        project_config=project_config, template=template, root_path=path, override=override
    )



"""The api module exposes the mloq functionality to the CLI."""
import copy
from pathlib import Path
from typing import Union

from mloq.config import Config
from mloq.directories import create_project_directories
from mloq.files import mlproject, OPEN_SOURCE_FILES, ROOT_PATH_FILES, SCRIPTS, setup_py, test_main
from mloq.requirements import setup_requirements
from mloq.templating import write_template
from mloq.workflows import setup_push_workflow


def requirements(
    project_config: Config,
    path: Union[Path, str],
    lint: bool = True,
    test: bool = True,
    install=None,
    override: bool = False,
):
    """Write requirements files and install them if requested."""
    if isinstance(install, (tuple, list)) and "all" in install or install == "all":
        install_reqs = True
        install_lint = True
        install_test = True
    elif isinstance(install, (tuple, list)):
        install_reqs = "reqs" in install or "requirements" in install
        install_lint = "lint" in install
        install_test = "test" in install
    else:
        install_reqs = False
        install_lint = False
        install_test = False
    options = project_config.get("requirements", ["empty"])
    setup_requirements(
        options=options,
        path=Path(path),
        test=test,
        lint=lint,
        install_test=install_test,
        install_reqs=install_reqs,
        install_lint=install_lint,
        override=override,
    )


def setup_project_files(path, template: Config, project_config: Config, override: bool = False):
    """Write the template for common repository config files."""
    path = Path(path)
    _template = copy.deepcopy(template)
    _template["project_name"] = _template["project_name"].replace("-", "_")
    project_name = _template["project_name"]
    create_project_directories(project_name=project_name, root_path=path, override=override)
    setup_root_files(
        template=_template, project_config=project_config, path=path, override=override
    )
    tests_path = path / project_name / "tests"
    write_template(test_main, template=_template, path=tests_path, override=override)


def setup_scripts(path: Union[str, Path], template: Config, override: bool = False):
    """Initialize CI scripts folder files."""
    path = Path(path)
    path = path if path.name == "scripts" else path / "scripts"
    for file in SCRIPTS:
        write_template(file, template=template, path=path, override=override)


def setup_root_files(
    path: Union[str, Path], template: Config, project_config: Config, override: bool = False
):
    """Initialize root folder files."""
    for file in ROOT_PATH_FILES:
        write_template(file, template=template, path=path, override=override)
    if project_config.get("mlflow"):
        write_template(mlproject, template=template, path=path, override=override)
    if project_config.get("open_source"):
        for file in OPEN_SOURCE_FILES:
            write_template(file, template=template, path=path, override=override)
    propietary_classif = "License :: Other/Proprietary License"
    license_classifiers = {"mit": "License :: OSI Approved :: MIT License"}
    setup_template = copy.deepcopy(template)
    license_classif = license_classifiers.get(template["license"], propietary_classif)
    setup_template["license_classifier"] = license_classif
    write_template(setup_py, template=setup_template, path=path, override=override)


def setup_repository(
    path: Union[str, Path], template: Config, project_config: Config, override: bool = False,
):
    """Initialize the project folder structure and all the filled in boilerplate files."""
    path = Path(path)
    setup_project_files(
        path=path, template=template, project_config=project_config, override=override
    )
    setup_push_workflow(
        project_config=project_config, path=path, template=template, override=override
    )
    requirements(project_config=project_config, path=path, test=True, lint=True, override=override)
    setup_scripts(template=template, path=path, override=override)



"""config contains all the logic needed to define the configuration of a new mloq project."""
from typing import Any, Dict, List, Set, Tuple, Union


Config = Dict[str, Union[None, str, Dict[str, Any], List[str]]]
Choices = Union[List[str], Tuple[str], Set[str]]



"""
This is mostly a copy paste from
https://github.com/pallets/click/blob/2fc486c880eda9fdb746ed8baa49416acab9ea6d/src/click/termui.py
Modified to allow prompt input that has a different color than the prompt text, while keeping
the color of the default prompt values the same as the prompt text color.
"""
import io

from click.exceptions import Abort, UsageError
from click.types import Choice, convert_type
from click.utils import echo, LazyFile


# The prompt functions to use.  The doc tools currently override these
# functions to customize how they work.
visible_prompt_func = input

_ansi_reset_all = "\033[0m"


def hidden_prompt_func(prompt):
    import getpass

    return getpass.getpass(prompt)


def _build_prompt(text, suffix, show_default=False, default=None, show_choices=True, type=None):
    prompt_ = text
    if type is not None and show_choices and isinstance(type, Choice):
        prompt_ += f" ({', '.join(map(str, type.choices))})"
    if default is not None and show_default:
        prompt_ = f"{prompt_} [{_format_default(default)}]"
    return f"{prompt_}{suffix}"


def _format_default(default):
    if isinstance(default, (io.IOBase, LazyFile)) and hasattr(default, "name"):
        return default.name

    return default


def prompt(
    text,
    default=None,
    hide_input=False,
    confirmation_prompt=False,
    type=None,
    value_proc=None,
    prompt_suffix=": ",
    show_default=True,
    err=False,
    show_choices=True,
):
    """Prompts a user for input.  This is a convenience function that can
    be used to prompt a user for input later.

    If the user aborts the input by sending a interrupt signal, this
    function will catch it and raise a :exc:`Abort` exception.

    .. versionadded:: 7.0
       Added the show_choices parameter.

    .. versionadded:: 6.0
       Added unicode support for cmd.exe on Windows.

    .. versionadded:: 4.0
       Added the `err` parameter.

    :param text: the text to show for the prompt.
    :param default: the default value to use if no input happens.  If this
                    is not given it will prompt until it's aborted.
    :param hide_input: if this is set to true then the input value will
                       be hidden.
    :param confirmation_prompt: asks for confirmation for the value.
    :param type: the type to use to check the value against.
    :param value_proc: if this parameter is provided it's a function that
                       is invoked instead of the type conversion to
                       convert a value.
    :param prompt_suffix: a suffix that should be added to the prompt.
    :param show_default: shows or hides the default value in the prompt.
    :param err: if set to true the file defaults to ``stderr`` instead of
                ``stdout``, the same as with echo.
    :param show_choices: Show or hide choices if the passed type is a Choice.
                         For example if type is a Choice of either day or week,
                         show_choices is true and text is "Group by" then the
                         prompt will be "Group by (day, week): ".
    """
    result = None

    def prompt_func(text):
        f = hidden_prompt_func if hide_input else visible_prompt_func
        try:
            # Write the prompt separately so that we get nice
            # coloring through colorama on Windows
            echo(f"{text}{_ansi_reset_all}", nl=False, err=err)
            return f("")
        except (KeyboardInterrupt, EOFError):
            # getpass doesn't print a newline if the user aborts input with ^C.
            # Allegedly this behavior is inherited from getpass(3).
            # A doc bug has been filed at https://bugs.python.org/issue24711
            if hide_input:
                echo(None, err=err)
            raise Abort()

    if value_proc is None:
        value_proc = convert_type(type, default)

    prompt_ = _build_prompt(text, prompt_suffix, show_default, default, show_choices, type)

    while 1:
        while 1:
            value = prompt_func(prompt_)
            if value:
                break
            elif default is not None:
                value = default
                break
        try:
            result = value_proc(value)
        except UsageError as e:
            if hide_input:
                echo("Error: the value you entered was invalid", err=err)
            else:
                echo(f"Error: {e.message}", err=err)  # noqa: B306
            continue
        if not confirmation_prompt:
            return result
        while 1:
            value2 = prompt_func("Repeat for confirmation: ")
            if value2:
                break
        if value == value2:
            return result
        echo("Error: the two entered values do not match", err=err)


def confirm(text, default=False, abort=False, prompt_suffix=": ", show_default=True, err=False):
    """Prompts for confirmation (yes/no question).

    If the user aborts the input by sending a interrupt signal this
    function will catch it and raise a :exc:`Abort` exception.

    .. versionadded:: 4.0
       Added the `err` parameter.

    :param text: the question to ask.
    :param default: the default for the prompt.
    :param abort: if this is set to `True` a negative answer aborts the
                  exception by raising :exc:`Abort`.
    :param prompt_suffix: a suffix that should be added to the prompt.
    :param show_default: shows or hides the default value in the prompt.
    :param err: if set to true the file defaults to ``stderr`` instead of
                ``stdout``, the same as with echo.
    """
    prompt_ = _build_prompt(text, prompt_suffix, show_default, "Y/n" if default else "y/N")
    while 1:
        try:
            # Write the prompt separately so that we get nice
            # coloring through colorama on Windows
            echo(f"{prompt_}{_ansi_reset_all}", nl=False, err=err)
            value = visible_prompt_func("").lower().strip()
        except (KeyboardInterrupt, EOFError):
            raise Abort()
        if value in ("y", "yes"):
            rv = True
        elif value in ("n", "no"):
            rv = False
        elif value == "":
            rv = default
        else:
            echo("Error: invalid input", err=err)
            continue
        break
    if abort and not rv:
        raise Abort()
    return rv



"""This module contains the functionality for parsing ad modifying the project configuration."""
from pathlib import Path
from typing import Dict, Optional, Union

from ruamel.yaml import load as yaml_load, Loader, YAML as RuamelYAML

from mloq.config import Config
from mloq.directories import copy_file
from mloq.files import file as new_file, mloq_yml
from mloq.requirements import require_cuda


def get_docker_python_version(template: Config) -> str:
    """Return the highest python version defined for the project."""
    max_version = list(sorted(template["python_versions"]))[-1]
    version = max_version.replace(".", "")
    return f"py{version}"


def get_docker_image(
    template: Config, project_config: Optional[Config] = None
) -> Union[str, None]:
    """
    Add to params the base docker container that will be used to define the project's container.

    If the dependencies require cuda the base image will be gpu friendly.
    """
    project_config = project_config or {}
    docker_is_false = project_config.get("docker") is not None and not project_config.get("docker")
    if docker_is_false:  # Project does not require Docker
        return "empty"
    # If docker image is defined, return current value even if it's an empty image
    if template.get("docker_image") is not None:
        return template["docker_image"]
    # Define the appropriate FragileTech docker container as base image
    v = get_docker_python_version(template)
    ubuntu_v = "ubuntu20.04" if v == "py38" else "ubuntu18.04"
    image = (
        f"fragiletech/{ubuntu_v}-cuda-11.0-{v}"
        if require_cuda(project_config)
        else f"fragiletech/{ubuntu_v}-base-{v}"
    )
    return image


def read_config(path: Union[Path, str]) -> Config:
    """Load the project configuration from the target path."""
    with open(path, "r") as config:
        params = yaml_load(config.read(), Loader)

    return params


def read_config_safe(path: Optional[Union[Path, str]]) -> Config:
    """
    Read the configuration file if it exists, otherwise return an empty config dict.

    Args:
        path: If it points to a file, it will read the file and return its content.
              If it points to a directory, it will try to read an mloq.yml file \
              present in that directory.

    Returns:
        Loaded configuration if a config file is present. Empty configuration otherwise.
    """
    if path is None:
        return load_empty_config()
    path = Path(path)
    path = path / mloq_yml.name if path.is_dir() else path
    try:
        return read_config(path)
    except FileNotFoundError:
        return load_empty_config()


def write_config(config: Config, path: Union[Path, str], safe: bool = False):
    """Write config in a yaml file."""
    yaml = RuamelYAML()
    yaml.indent(sequence=4, offset=2)
    if safe:
        path = Path(path)
        path = path / mloq_yml.name if path.is_dir() else path
    with open(path, "w") as f:
        yaml.dump(config, f)


def load_empty_config() -> Dict[str, Dict[str, None]]:
    """Return a dictionary containing all the MLOQ config values set to None."""
    with open(mloq_yml.src, "r") as config:
        empty_config = yaml_load(config.read(), Loader)
    return empty_config


def write_empty_config(path: Union[str, Path], override: bool = False, filename: str = None):
    """Write an empty config file to the target path."""
    repo_file = (
        mloq_yml if filename is None else new_file(mloq_yml.src, mloq_yml.src.parent, filename)
    )
    copy_file(repo_file, Path(path), override)



"""Command line interface for mloq."""
import os

import click


override_opt = click.option(
    "--override/--no-override",
    "-o/ ",
    default=False,
    show_default=True,
    help="If True overwrite existing files. If False ignore files that already exist.",
)
config_file_opt = click.option(
    "--filename",
    "-f",
    "config_file",
    default=None,
    show_default=True,
    help="Name of the repository config file",
    type=click.Path(exists=True, file_okay=True, dir_okay=True, resolve_path=True),
)


output_path_arg = click.argument(
    "output", type=click.Path(exists=True, file_okay=False, dir_okay=True, resolve_path=True)
)

interactive_opt = click.option(
    "--interactive/--no-interactive",
    "-i/ ",
    default=False,
    show_default=True,
    help="If True the configuration values will be defined interactively on the command line.",
)


def _parse_env(config_file, override, interactive):
    """Parse environment variables defining command options."""
    config_file = os.getenv("MLOQ_CONFIG_FILE", config_file)
    override = os.getenv("MLOQ_OVERRIDE", override)
    interactive = os.getenv("MLOQ_INTERACTIVE", interactive)
    return config_file, override, interactive


@click.group()
def cli():
    """Command line interface for ML Ops Quickstart."""
    pass


@cli.command()
@override_opt
@interactive_opt
@config_file_opt
@output_path_arg
def setup(config_file, output, override: bool, interactive: bool):
    from mloq.cli.setup_cmd import setup_cmd

    config_file, override, interactive = _parse_env(config_file, override, interactive)
    setup_cmd(config_file, output, override, interactive)



import copy
import os

import pytest

from mloq.config.generation import generate_project_config, generate_template


@pytest.fixture()
def project_config():
    test_project_config = {
        "open_source": True,
        "docker": True,
        "ci": "python",
        "mlflow": True,
        "requirements": ["torch"],
    }
    return test_project_config


@pytest.fixture()
def template():
    test_template = {
        "project_name": "test_project",
        "default_branch": "master",
        "owner": "owner_test",
        "author": "owner_test",
        "email": "owner@mail.com",
        "copyright_holder": "owner_test",
        "project_url": "https://github.com/owner_test/test_project",
        "bot_name": "owner_test",
        "bot_email": "owner@mail.com",
        "license": "MIT",
        "description": "test description of the project",
        "python_versions": ["3.6", "3.7", "3.8", "3.9"],
        "docker_image": "fragiletech/ubuntu18.04-cuda-11.0-py39",
    }
    return test_template


def compare_dicts(a, b):
    for (k1, v1), (k2, v2) in zip(a.items(), b.items()):
        assert k1 == k2
        if not isinstance(v1, list):
            assert v1 == v2
        else:
            for x1, x2 in zip(v1, v2):
                assert x1 == x2


def test_generate_config(project_config):
    in_config = copy.deepcopy(project_config)
    in_config["requirements"] = None
    os.environ["MLOQ_REQUIREMENTS"] = "torch"
    new_template = generate_project_config(project_config=in_config, interactive=False)
    del os.environ["MLOQ_REQUIREMENTS"]
    compare_dicts(project_config, new_template)


def test_generate_template(template, project_config):
    in_template = copy.deepcopy(template)
    in_template["project_name"] = None
    os.environ["MLOQ_PROJECT_NAME"] = "test_project"
    new_template = generate_template(
        template=in_template, project_config=project_config, interactive=False
    )
    del os.environ["MLOQ_PROJECT_NAME"]
    compare_dicts(template, new_template)
